// Copyright (c) 2015, Anders Holmgren. All rights reserved. Use of this source code
// is governed by a BSD-style license that can be found in the LICENSE file.

library stuff.json;

import 'package:path/path.dart' as path;
import 'package:uri/uri.dart';
import 'package:option/option.dart';

final _p = path.url;

abstract class Jsonable {
  toJson();
}

JsonBuilder get buildJson => new JsonBuilder();
JsonParser parseJson(Map j, {bool consumeMap: false}) =>
    new JsonParser(j, consumeMap);

class JsonBuilder {
  final bool _stringEmpties;
  JsonBuilder({bool stringEmpties: true}) : this._stringEmpties = stringEmpties;

  final Map json = {};

//  void addObject(String fieldName, o) {
//    if (o != null) {
//      json[fieldName] = o.toJson();
//    }
//  }

  void add(String fieldName, v, [transform(v)]) {
    if (v != null) {
      final transformed = _transformValue(v, transform);
      if (transformed != null) {
        json[fieldName] = transformed;
      }
    }
  }

  void addAll(Map map) {
    json.addAll(map);
  }

  _transformValue(value, [transform(v)]) {
    if (transform != null) {
      return transform(value);
    }
    if (value is Jsonable) {
      return value.toJson();
    }
    if (value is Map) {
      final result = {};
      (value as Map).forEach((k, v) {
        final transformed = _transformValue(v);
        if (transformed != null) {
          result[k] = transformed;
        }
      });
      return result.isNotEmpty || !_stringEmpties ? result : null;
    }
    if (value is Iterable) {
      final list =
          (value as Iterable).map((v) => _transformValue(v, null)).toList();
      return list.isNotEmpty || !_stringEmpties ? list : null;
    }
    if (value is RegExp) {
      return value.pattern;
    }
    if (value is UriTemplate) {
      return value.template;
    }
    if (value is DateTime) {
      return value.toIso8601String();
    }
    if (value is bool || value is num) {
      return value;
    }
    if (value is Option) {
      return value is Some ? _transformValue(value.get()) : null;
    }
    return value.toString();
  }
}

_identity(v) => v;

class JsonParser {
  final Map _json;
  final bool _consumeMap;

  JsonParser(Map json, bool consumeMap)
      : this._json = consumeMap ? new Map.from(json) : json,
        this._consumeMap = consumeMap;

  List list(String fieldName, [create(i) = _identity]) {
    final List l = _getField(fieldName);
    return l != null ? l.map(create).toList(growable: false) : [];
  }

  single(String fieldName, [create(i) = _identity]) {
    final j = _getField(fieldName);
    return j != null ? create(j) : null;
  }

  Map mapValues(String fieldName, [create(i) = _identity]) {
    final Map m = _getField(fieldName);

    if (m == null) {
      return {};
    }

    Map result = {};
    m.forEach((k, v) {
      result[k] = create(v);
    });

    return result;
  }

  _getField(String fieldName) =>
      _consumeMap ? _json.remove(fieldName) : _json[fieldName];

  Map get unconsumed {
    if (!_consumeMap) {
      throw new StateError('unconsumed called on non consuming parser');
    }

    return _json;
  }
}
